/******************************************************************************

    main.c

    UVC Webcamera Driver for PIC24FJ64GB002/PIC24FJ256GB106/PIC32MX220F032B/PIC32MX695F512H

    Driver test by ELECOM UVC Webcamera UCAM-DLK130T.

    BUFFALO UVC Webcam BSW13K08H doesn't work. It used eSP368
  usb camera module by Etron Technology. Other module eSP268
  doesn't work, too. These camera modules don't good work in
  Full Speed.


*** PIC24FJ64GB002 pin assign

  EPSON LCDC SD13781 SPI INTERFACE
    PIC24FJ64B002                       LCDC
	RB15(26) SPI CSS# (Chip Select) --- J4-27 CS#_SCS#
	RB14(25) SPI Clock              --- J4-28 WR#_SCK
	RB9 (18) SPI MOSI (Data Out)    --- J4-29 DB0_SI
	RB8 (17) SPI MISO (Data In)     --- J4-30 DB1_SO

  UART2
	RA0/RP5 (2) --- U2TX
	RA1/RP6 (3) --- U2RX

*** PIC24FJ256GB106 pin assign

  EPSON LCDC SD13781 INDIRECT 8 bit mode INTERFACE
    PIC24USB board LCDC 
	RG9@CN1-11 --- J4-6  P/C#_AB1
	RG8 CN1-9  --- J4-28 WR#_SCK
	RG7 CN1-8  --- J4-25 RD#
	RG6 CN1-7  --- J4-27 CS#_SCS#

 	RE7 CN1-6  --- J4-36 DB7
 	RE6 CN1-5  --- J4-35 DB6
	RE5 CN1-4  --- J4-34 DB5
	RE4 CN2-3  --- J4-33 DB4
	RE3 CN2-4  --- J4-32 DB3
	RE2 CN2-5  --- J4-31 DB2
	RE1 CN2-6  --- J4-30 DB1_SO
	RE0 CN2-7  --- J4-29 DB0_SI

  UART2
	RF4/RP10 CN2-26 --- U2TX
	RF5/RP17 CN2-25 --- U2RX


*** PIC32MX220F03B pin assign

  EPSON LCDC SD13781 SPI INTERFACE
    PIC32MX220F032B                          LCDC
	RB15      (26) SPI CSS# (Chip Select) --- J4-27 CS#_SCS#
	RB14/SCK1 (25) SPI Clock              --- J4-28 WR#_SCK
	RPB13/RB13(24) SPI MOSI (Data Out)    --- J4-29 DB0_SI
	RPB8/RB8  (17) SPI MISO (Data In)     --- J4-30 DB1_SO

  UART2
	RPA1/RA1 ( 3) --- U2RX
	RPA3/RA3 (10) --- U2TX

*** PIC32MX695F512H pin assign

  EPSON LCDC SD13781 SPI INTERFACE
    PIC32MX220F032B                          LCDC
	RD4      (52) SPI CSS# (Chip Select) --- J4-27 CS#_SCS#
	SCK3     (49) SPI Clock              --- J4-28 WR#_SCK
	SDO3     (51) SPI MOSI (Data Out)    --- J4-29 DB0_SI
	SDI3     (50) SPI MISO (Data In)     --- J4-30 DB1_SO

  UART2
	U2RX     (31)
	U2TX     (32)

  USB
	D-       (36)
	D+       (37)

******************************************************************************/

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "usb.h"
#include "usb_host_uvc.h"
#include "uart2.h"
#include "s1d13781.h"
#include "pic_delay.h"

/*------------------------------------------------------------------------------
	Configuration for PIC24F/PIC32MX 
------------------------------------------------------------------------------*/

#if defined(__PIC24FJ64GB002__)
    // PIC24FJ64GB002 
    _CONFIG1(WDTPS_PS1 & FWPSA_PR32 & WINDIS_OFF & FWDTEN_OFF & ICS_PGx1 & GWRP_OFF & GCP_OFF & JTAGEN_OFF)
    _CONFIG2(POSCMOD_NONE & I2C1SEL_PRI & IOL1WAY_OFF & OSCIOFNC_ON & FCKSM_CSDCMD & FNOSC_FRCPLL & PLL96MHZ_ON & PLLDIV_DIV2 & IESO_OFF)
    _CONFIG3(WPFP_WPFP0 & SOSCSEL_IO & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
    //_CONFIG3(WPFP_WPFP0 & SOSCSEL_SOSC & WUTSEL_LEG & WPDIS_WPDIS & WPCFG_WPCFGDIS & WPEND_WPENDMEM)
    _CONFIG4(DSWDTPS_DSWDTPS3 & DSWDTOSC_LPRC & RTCOSC_SOSC & DSBOREN_OFF & DSWDTEN_OFF)
#elif defined(__PIC24FJ256GB106__)
    // PIC24FJ256GB106 
    _CONFIG1( JTAGEN_OFF & GCP_OFF & GWRP_OFF & FWDTEN_OFF & ICS_PGx2 & BKBUG_OFF & COE_OFF ) 
    _CONFIG2( PLL_96MHZ_ON & IESO_OFF & FCKSM_CSDCMD & OSCIOFNC_OFF & POSCMOD_XT & FNOSC_PRIPLL & PLLDIV_DIV4 & IOL1WAY_OFF & DISUVREG_OFF) 
    _CONFIG3( WPEND_WPENDMEM & WPCFG_WPCFGDIS & WPDIS_WPDIS )
#elif defined(__32MX220F032B__)
	#include "p32mx220f_config.h"
#elif defined(__32MX695F512H__)
	#include "p32mx695f_config.h"
#else
	#error  "Undefined processor"
#endif


/*------------------------------------------------------------------------------
	definition
------------------------------------------------------------------------------*/

#define	putstr(x)	UART2PrintString(x)
#define	inkey()		UART2IsPressed()
#define	getch()		UART2GetChar()
#define	puthex(x)	UART2PutHex(x)
#define	putch(x)	UART2PutChar(x)

typedef enum
{
	CAMERA_DETACHED = 0,
	CAMERA_ATTACHED,
} CAMERA_STATE;


/*------------------------------------------------------------------------------
	function prototype
------------------------------------------------------------------------------*/

void	InitializeSystem (void);
void	getstr(char *buf);
void	putdata(BYTE *data, BYTE len, WORD type);
void	yuv2rgb_mono(BYTE *dst, BYTE *src, WORD length);
void	yuv2rgb_color(BYTE *dst, BYTE *src, WORD length);


/*------------------------------------------------------------------------------
	global variable
------------------------------------------------------------------------------*/

CAMERA_STATE camera_state;
char	*request_name[] = {
	"Current: ",
	"Minimum: ",
	"Maximum: ",
	"Resolution: ",
	"Length: ",
	"Info: ",
	"Default: ",
};


/*==============================================================================
	main
===============================================================================*/

int main ( void )
{
	int		c, n, m;
	char	buf[64];

	BYTE	rc;
	BYTE	control, len;
	BYTE	data[64];
	WORD	type;
	BYTE 	maxformat, format, maxframe, set_flg;
	UVC_FRAME_INFO	*frame_info;
	WORD 	width, height;
	DWORD 	frame_size;
	UVC_DATA_BUFFER	*dp;
	long	addr;
	BYTE	vram_sw;

    // Initialize the processor and peripherals.
    InitializeSystem();

	// opening message
    putstr("\r\n*** UVC webcamera driver test program.\r\n" );

	// initilze status
	camera_state = CAMERA_DETACHED;
	set_flg = 0;

	for(;;){
restart:
		// wait for camera atached
		putstr("  wait for Camera Attached.\r\n");
		while(camera_state != CAMERA_ATTACHED){
			USBHostTasks();
		}

		// clear key buffer
		while(inkey()){
			getch();
		}

		for(;;){
			putstr("\r\n***** UVC Camera Driver Test Menu ************\r\n");
			putstr("d:display Test Bar on LCD.\r\n");
			putstr("c:List Camera Control current value.\r\n");
			putstr("f:List supported Format.\r\n");
			putstr("i:List supported Image Resolution.\r\n");
			putstr("p:Set capture image Format & Resolution.\r\n");
			putstr("e:execute frame capturing.\r\n"); 
			putstr("=>");

			while(inkey() == 0){
				USBHostTasks();
				if(camera_state == CAMERA_DETACHED){
					goto restart;
				}
			}

			c = getch();
			if(c >= ' ' && c <= 0x7e){
				putch(c);
				putstr("\r\n");
			}

			switch(c){

			// Test Bar display on color LCD.
			case 'd':
				vram_testbar_write(MAIN_LAYER);
				break;

			// Listng Camera Control value.
			case 'c':
				for(control = UVC_APL_SCANNING; control < UVC_APL_PROBE; control++){

					// first request is GET_CUR
					rc = USBHostUVCGetControlValue(control, GET_CUR, data, &len, &type);
					// if control isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print control name
					putstr(USBHostUVCGetControlName(control)); putstr(":\r\n");
					// print current setting
					putstr("\tCur "); putdata(data, len, type); putstr("\r\n");

					// 2nd request is GET_MIN
					rc = USBHostUVCGetControlValue(control, GET_MIN, data, &len, &type);
					// if GET_MIN request isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print minimum setting
					putstr("\tMin "); putdata(data, len, type); putstr("\r\n");

					// 3rd request is GET_MAX
					rc = USBHostUVCGetControlValue(control, GET_MAX, data, &len, &type);
					// if GET_MAX request isn't supported, loop continue to next control. 
					if(rc != UVC_SUCCESS) continue;

					// print maximum setting
					putstr("\tMax "); putdata(data, len, type); putstr("\r\n");

				} // for(control)
				break;

			// Listing Support Image Format
			case 'f':
				maxformat = USBHostUVCGetFormatCount();
				for(n = 0; n < maxformat; n++){
					format = USBHostUVCGetFormat(n);
					sprintf(buf, "Supported Format (Number is %d): ", n); putstr(buf);
					if(format == FMT_YUY2) putstr("YUY2");
					if(format == FMT_NV12) putstr("NV12");
					if(format == FMT_MJPEG) putstr("MJPEG");
					if(format == FMT_MPEG2TS) putstr("MPEG2 TS");
					if(format == FMT_DV) putstr("DV");
					putstr("\r\n");
				}
				break;

			// Listing Support Frame Resolution
			case 'i':
				maxformat = USBHostUVCGetFormatCount();
				for(n = 0; n < maxformat; n++){
					format = USBHostUVCGetFormat(n);
					sprintf(buf, "Supported Format (Number is %d): ", n); putstr(buf);
					if(format == FMT_YUY2) putstr("YUY2");
					if(format == FMT_NV12) putstr("NV12");
					if(format == FMT_MJPEG) putstr("MJPEG");
					if(format == FMT_MPEG2TS) putstr("MPEG2 TS");
					if(format == FMT_DV) putstr("DV");
					putstr("\r\n");

					frame_info = USBHostUVCGetFrame(n, &maxframe);
					putstr("  Resolution:\r\n\t");
					for(m = 0; m < maxframe; m++){
						width = frame_info[m].wWidth;
						height = frame_info[m].wHeight;
						sprintf(buf, "%dx%d ", width, height); putstr(buf);
						if(m % 3 == 2) putstr("\r\n\t");
					}
					putstr("\r\n");
				}
				break;

			// Set frame capture format and reslution
			case 'p':
				putstr("Format Number=>");
				getstr(buf); n = atoi(buf);
				if(n > USBHostUVCGetFormatCount()){
					putstr("Format Number error!\r\n");
					break;
				}
				putstr("Resolution width=>");
				getstr(buf); width = atoi(buf);
				putstr("Resolution height=>");
				getstr(buf); height = atoi(buf);
				if(width > 640 || height > 480){
					putstr("Resolution Unsupported!\r\n");
					break;
				}

				frame_size = (DWORD)width * (DWORD)height * 2L;
				format = USBHostUVCGetFormat(n);

				rc = USBHostUVCSetFrame(n, width, height);
				if(rc == UVC_UNSUPPORTED_FORMAT){
					putstr("Format Unspported!\r\n");
				}else if(rc == UVC_UNSUPPORTED_RESOLUTION){
					putstr("Resolution Unspported!\r\n");
				}else if(rc != UVC_SUCCESS){
					putstr("Set Error!\r\n");
				}else{
					putstr("Format and Resolution, Set ok.\r\n");
					sprintf(buf, "frame_size=%ld\r\n", frame_size); putstr(buf);
					set_flg = 1;
				}
				break;


			// execute frame capturing
			case 'e':
				if(!set_flg){
					putstr("Format and Resolution was no set.\r\n");
					break;
				}

				putstr("Capture Start!\r\n");

				// start frame capturing
				if(USBHostUVCStartFrameCapturing() == UVC_FAIL){
					putstr(" ... found error!\r\n");
					break;
				}

				// LCD layer setting
				// now layer is main layer.				
				vram_sw = PIP_LAYER;
				addr = PIP_VRAM_ADR;

				// capture loop
				for(;;){
					if((dp = USBHostUVCGetFrameCapturing()) == NULL) break;
					if(dp->length == 0){	// found Header only, End of Frame
						*(dp->valid) = 0;	// clear valid flag

						// Display Layer
						//  when now write is PIP, display PIP layer
						if(vram_sw == MAIN_LAYER){
							select_display_layer(MAIN_LAYER);
							addr = PIP_VRAM_ADR;
							vram_sw = PIP_LAYER;
						}else{
							select_display_layer(PIP_LAYER);
							addr = MAIN_VRAM_ADR;
							vram_sw = MAIN_LAYER;
						}
						continue;
					}

					// Display YUYV image format,
					//   skip display other format.
					if(format == FMT_YUY2){

						// convert yuyv to RGB565
						#if defined(__PIC24FJ64GB002__)
							yuv2rgb_mono(dp->buffer, dp->buffer, dp->length);
						#else
							yuv2rgb_color(dp->buffer, dp->buffer, dp->length);
						#endif
	
						// display to LCD
						write_byte_block(addr, dp->buffer, dp->length);
						addr += dp->length;
					}

					// clear valid flag
					*(dp->valid) = 0;

					// check user stop
					if(inkey()){
						getch();
						break;
					}
				} // capture loop

				putstr("Capture Stop!\r\n");


				if(camera_state == CAMERA_DETACHED){
					goto restart;
				}

				// Stop Frame Capture
				USBHostUVCStopFrameCapturing();

				set_flg = 0;	// clear resolution set flag
				break;

			default:
				break;

			} // switch(c)

		} // for (menu loop)

	}// for (main loop)

    return 0;

} // main 


/*------------------------------------------------------------------------
	InitializeSystem()
-----------------------------------------------------------------------*/

void InitializeSystem ( void )
{
#if defined(__PIC24FJ64GB002__)
	unsigned int pll_startup_counter;

	// Fosc(system clock) 8MHz postscale = 1:1 x4 = 32MHz, Fcy = Fosc/2
	CLKDIV = 0;

	// PLL clock start up
	pll_startup_counter = 600;
	CLKDIVbits.PLLEN = 1;
	while(pll_startup_counter--);

	//--- setting interrupt ---------------------------------------
	INTCON1 = 0;			// NSTDIS=0 nesting ok

	//--- setting I/O port ----------------------------------------
	AD1PCFG = 0xffff;	// Analog disable
	LATA = 0x0000; TRISA = 0xffff;	// RA are default input.
	LATB = 0x0000; TRISB = 0xffff;	// RB are default input.

	TRISAbits.TRISA0 = 0;	// RA0(UxTX) is output
	TRISAbits.TRISA1 = 1;	// RA1(UxRX) is input
	LATAbits.LATA0 = 1;		// RA0(UxTX) to level high

	//--- remapping I/O pin for peripheral ------------------------
	//unlock registers
	__builtin_write_OSCCONL(OSCCON & 0xBF);

	// Assign UART1	
	//RPOR2bits.RP5R = 3;		// RA0/RP5(pin 2)->UART1 TX
	//RPINR18bits.U1RXR = 6;	// RA1/RP6(pin 3)->UART1 RX

	// Assign UART2	
	RPOR2bits.RP5R = 5;		// RA0/RP5(pin 2)->UART2 TX
	RPINR19bits.U2RXR = 6;	// RA1/RP6(pin 3)->UART2 RX

	// Assign SPI1  (for PIC24FJ64GB002)
	RPOR4bits.RP9R = 7;      // RP9(RB9)-> SPI1 Data Output(7)
	RPINR20bits.SDI1R = 8;   // SPI1 Data Input <- RP8(RB8)
	RPOR7bits.RP14R = 8;     // RP14(RB14)-> SPI1 Clock Outpur(8)

	TRISBbits.TRISB15 = 0;	// Chip Select   I/O out
	TRISBbits.TRISB14 = 0;	// SPI Clock Out I/O out
	TRISBbits.TRISB9 = 0;	// SPI Data Out  I/O out
	TRISBbits.TRISB8 = 1;	// SPI Data In   I/O in

	//lock registers
	__builtin_write_OSCCONL(OSCCON | 0x40);

	//--- setting SPI for S1D13781 with PIC24FJ64GB002 ------------
	// Slave Select(chip select) 
	LATBbits.LATB15 = 1;		// RB15 SS#, to HIGH Level(select)
	// for S1D13781 LCD Controler
	// 	SCLOCK normal low or High
	// 	DOUT out fall edg, S1d13781 read rising edge
	// 	DIN  rising IN, S1d13781 out fall edge
	// 	Normal Low CKP = 0, CKE=1 / Normal High CKP = 1, CKE = 0
	//	SDI data sampling, rising edge, SMP = 0
	SPI1STAT = 0;	// SPI1 is disable
	SPI1CON1bits.DISSCK = 0;
	SPI1CON1bits.DISSDO = 0;
	SPI1CON1bits.MODE16 = 0; // 8bits
	SPI1CON1bits.SMP = 0;
	SPI1CON1bits.CKE = 0;
	SPI1CON1bits.SSEN = 0;	// don't slave
	SPI1CON1bits.CKP = 1;
	SPI1CON1bits.MSTEN = 1;	// master mode
	//SPI1CON1bits.SPRE = 0b110;	// second prescale 2:1
	//SPI1CON1bits.PPRE = 0b11;	// first prescale 1:1
	// 16M/2=8Mhz

	SPI1CON1bits.SPRE = 0b111;	// second prescale 1:1
	SPI1CON1bits.PPRE = 0b11;	// first prescale 1:1
	// 16M/1=16Mhz

	SPI1CON2 = 0;
	SPI1STATbits.SPIEN = 1;		// start SPI1

#elif defined(__PIC24FJ256GB106__)

	// Fosc(system clock) 8MHz postscale = 1:1 x4 = 32MHz, Fcy = Fosc/2
	CLKDIV = 0;

	//--- setting interrupt ---------------------------------------
	INTCON1 = 0;			// NSTDIS=0 nesting ok

	//--- setting I/O port ----------------------------------------
	AD1PCFG = 0xffff;	// Analog disable
	LATB = 0x0000; TRISB = 0xffff;	// RB are default input.
	LATC = 0x0000; TRISC = 0xffff;	// RC are default input.
	LATD = 0x0000; TRISD = 0xffff;	// RD are default input.
	LATE = 0x0000; TRISE = 0xffff;	// RE are default input.
	LATF = 0x0000; TRISF = 0xffff;	// RF are default input.
	LATG = 0x0000; TRISG = 0xffff;	// RG are default input.

	//--- remapping I/O pin for peripheral ------------------------
	//unlock registers
	__builtin_write_OSCCONL(OSCCON & 0xBF);

	// Assign UART1	
	//RPOR5bits.RP10R = 3;		// RF4/RP10 <- UART1 TX(3)
	//RPINR18bits.U1RXR = 17;	// UART1RX  <- RF5/RP17

	// Assign UART2	
	RPOR5bits.RP10R = 5;	// RF4/RP10 <- UART2 TX(5)
	RPINR19bits.U2RXR = 17;	// UART2RX  <- RF5/RP17

	TRISFbits.TRISF4 = 0;	// RF4(UxTX) is output
	TRISFbits.TRISF5 = 1;	// RF5(UxRX) is input
	LATFbits.LATF4 = 1;		// RF4(UxTX) to level high

	//lock registers
	__builtin_write_OSCCONL(OSCCON | 0x40);

	//--- setting I/O for S1D13781 with PIC24FJ256GB106
	// indirect 8bit interface control
	LATG = (0b0111) << 6;	// P/C#=0,WR#=1,RD#=1,CS#=1
	TRISG = 0;				// control signal out
	LATE = 0; TRISE = 0; // DB0-DB7 to output

#elif defined(__32MX220F032B__)
	int	value;

	// Set System Config
	//SYSTEMConfigPerformance(GetSystemClock());

	value = SYSTEMConfigWaitStatesAndPB( GetSystemClock() );

	// Enable the cache for the best performance	
	CheKseg0CacheOn();

	INTEnableSystemMultiVectoredInt();

	value = OSCCON;
	while (!(value & 0x00000020)){
		value = OSCCON;    // Wait for PLL lock to stabilize	
	}

	INTEnableInterrupts();


	// ADC input off
	ANSELA = 0;
	ANSELB = 0;

	// Assign UART2	
	U2RXRbits.U2RXR = 0;	// UART2 RX <- RPA1(pin 3)[0]
	RPA3Rbits.RPA3R = 2;	// RPA3(pin10)<- UART2 TX [2]

	// rs initialize
	UART2Init();
	//rs_initialize();

	// Assign SPI1  (for PIC32MX220F032B)
	// SPI SSN#     RB15(26)  Chip Select
	// SPI Clock    RB14(25)
	// SPI Data Out RB13(24)  MOSI
	// SPI Data In  RB8 (17)  MISO

	SDI1Rbits.SDI1R = 4;   // SPI1 Data Input <- RPB8/RB8(17)
	RPB13Rbits.RPB13R = 3; // RPB13/RB13-> SPI1 Data Output(24)
	// TRISBbits.TRISB15 = 0;	Chip Select   I/O out (26)
	TRISBCLR = 0x8000;		
	// RPB14 is SCK1 (25)

	//--- setting SPI for S1D13781 with PIC24FJ64GB002 ------------
	// Slave Select(chip select) 
	//LATBbits.LATB15 = 1;		// RB15 SS#, to HIGH Level(select)
	LATBSET = 0x8000;

	// for S1D13781 LCD Controler
	// 	SCLOCK normal low or High
	// 	DOUT out fall edg, S1d13781 read rising edge
	// 	DIN  rising IN, S1d13781 out fall edge
	// 	Normal Low CKP = 0, CKE=1 / Normal High CKP = 1, CKE = 0
	//	SDI data sampling, rising edge, SMP = 0
/***
	SPI1CONbits.FRMEN = 0;		// Framed SPI disable
	SPI1CONbits.FRMSYNC = 0;	// Frame Sync pulse output(master mode)
	SPI1CONbits.FRMPOL = 0;		// Frame pulse is active-low
	SPI1CONbits.MSSEN = 0;		// Slave select SPI support is disabled
	SPI1CONbits.FRMSYPW = 0;	// Frame sync pulse is one clock wide
	SPI1CONbits.FRMCNT = 0b000;	// Generate a frame sync pulse on every data character
	SPI1CONbits.MCLKSEL = 0;	// PBCLK is used by the Baud Rate Generator
	SPI1CONbits.SPIFE = 0;		// Frame synchronization pulse precedes the first bit clock
	SPI1CONbits.ENHBUF = 0;		// Enhanced Buffer mode is disabled
	SPI1CONbits.ON = 0;			// SPI Peripheral is disabled
	SPI1CONbits.SIDL = 0;		// Continue operation in Idle mode
	SPI1CONbits.DISSDO = 0;		// SDOx pin is controlled by the SPI module
	SPI1CONbits.MODE32 = 0;		// 
	SPI1CONbits.MODE16 = 0;		// Communication 8-bit when SPI1CON2bits.AUDEN=0
	SPI1CONbits.SMP = 0;		// Input data sampled at middle of data output time
	SPI1CONbits.CKE = 0;		// Serial output data changes on transition from Idle clock state to active clock state (see CKP bit)
	SPI1CONbits.SSEN = 0;		// SSx pin not used for Slave mode, pin controlled by port function
	SPI1CONbits.CKP = 1;		// Idle state for clock is a low level; active state is a high level
	SPI1CONbits.MSTEN = 1;		// Master mode
	SPI1CONbits.DISSDI = 0;		// SDI pin is controlled by the SPI module
	SPI1CONbits.STXISEL = 0b00;	// Interrupt is generated when the last transfer is shifted out of SPISR and transmit operations are complete
	SPI1CONbits.SRXISEL = 0b00;	// Interrupt is generated when the last word in the receive buffer is read (i.e., buffer is empty)

	SPI1CON2bits.SPISGNEXT = 0;	// Data from RX FIFO is not sign extened
	SPI1CON2bits.FRMERREN = 0;	// Frame Error does not generate error events
	SPI1CON2bits.SPIROVEN = 0;	// Receive overflow does not generate error events
	SPI1CON2bits.SPITUREN = 0;	// Transmit Underrun Does Not Generates Error Events
	SPI1CON2bits.IGNROV = 0;	// A ROV is a critical error which stop SPI operation
	SPI1CON2bits.IGNTUR	= 0;	// A TUR is a critical error which stop SPI operation
	SPI1CON2bits.AUDEN = 0;		// Audio protocol disabled
	SPI1CON2bits.AUDMONO = 0;	// Audio data is stereo
	SPI1CON2bits.AUDMOD = 0b00;	// Audio protocol mode is I2S mode

	SPI1STATbits.SPIROV = 0;	// cleared Receive Overflow Flag bit. 
	SPI1CONbits.ON = 1;	// SPI Peripheral is enabled

***/
	SPI1CON = 0;
	SPI1CON2 = 0;
	SPI1BRG = 1;			// SCK1 = 80MHz/2/(SPI1BRG+1) = 20MHz
	SPI1CONSET = 0x8060;	// ON=1, CKP=1, MSTEN=1
	SPI1STATCLR = 0x40;		// clear SPIROV


#elif defined(__32MX695F512H__)

	int	value;

	// Set System Config
	//SYSTEMConfigPerformance(GetSystemClock());

	value = SYSTEMConfigWaitStatesAndPB( GetSystemClock() );

	// Enable the cache for the best performance	
	CheKseg0CacheOn();

	INTEnableSystemMultiVectoredInt();

	value = OSCCON;
	while (!(value & 0x00000020)){
		value = OSCCON;    // Wait for PLL lock to stabilize	
	}

	INTEnableInterrupts();

	// ADC input off
	AD1PCFG = 0xffff;

	// rs initialize
	UART2Init();
	//rs_initialize();

	// Assign SPI3 (for PIC32MX695F512H)
	// SPI3 Clock         (49) SCK
	// SPI3 Data In       (50) MISO
	// SPI3 Data Out      (51) MOSI
	// Slave Select# RD4  (52) Chip Select

	// TRISDbits.TRISD4 = 0; Chip Select I/O out (52)
	TRISDCLR = 0x0010;

	//--- setting SPI for S1D13781 with PIC24FJ64GB002 ------------
	// Slave Select(chip select) 
	//LATDbits.LATD4 = 1;	// RD4 SS#, to HIGH Level(deselect)
	LATDSET = 0x0010;

	// for S1D13781 LCD Controler
	// 	SCLOCK normal low or High
	// 	DOUT out fall edg, S1d13781 read rising edge
	// 	DIN  rising IN, S1d13781 out fall edge
	// 	Normal Low CKP = 0, CKE=1 / Normal High CKP = 1, CKE = 0
	//	SDI data sampling, rising edge, SMP = 0
/***
	SPI3CONbits.FRMEN = 0;		// Framed SPI disable
	SPI3CONbits.FRMSYNC = 0;	// Frame Sync pulse output(master mode)
	SPI3CONbits.FRMPOL = 0;		// Frame pulse is active-low
	SPI3CONbits.MSSEN = 0;		// Slave select SPI support is disabled
	SPI3CONbits.FRMSYPW = 0;	// Frame sync pulse is one clock wide
	SPI3CONbits.FRMCNT = 0b000;	// Generate a frame sync pulse on every data character
	SPI3CONbits.MCLKSEL = 0;	// PBCLK is used by the Baud Rate Generator
	SPI3CONbits.SPIFE = 0;		// Frame synchronization pulse precedes the first bit clock
	SPI3CONbits.ENHBUF = 0;		// Enhanced Buffer mode is disabled
	SPI3CONbits.ON = 0;			// SPI Peripheral is disabled
	SPI3CONbits.SIDL = 0;		// Continue operation in Idle mode
	SPI3CONbits.DISSDO = 0;		// SDOx pin is controlled by the SPI module
	SPI3CONbits.MODE32 = 0;		// 
	SPI3CONbits.MODE16 = 0;		// Communication 8-bit when SPI1CON2bits.AUDEN=0
	SPI3CONbits.SMP = 0;		// Input data sampled at middle of data output time
	SPI3CONbits.CKE = 0;		// Serial output data changes on transition from Idle clock state to active clock state (see CKP bit)
	SPI3CONbits.SSEN = 0;		// SSx pin not used for Slave mode, pin controlled by port function
	SPI3CONbits.CKP = 1;		// Idle state for clock is a low level; active state is a high level
	SPI3CONbits.MSTEN = 1;		// Master mode
	SPI3CONbits.DISSDI = 0;		// SDI pin is controlled by the SPI module
	SPI3CONbits.STXISEL = 0b00;	// Interrupt is generated when the last transfer is shifted out of SPISR and transmit operations are complete
	SPI3CONbits.SRXISEL = 0b00;	// Interrupt is generated when the last word in the receive buffer is read (i.e., buffer is empty)

	SPI3CON2bits.SPISGNEXT = 0;	// Data from RX FIFO is not sign extened
	SPI3CON2bits.FRMERREN = 0;	// Frame Error does not generate error events
	SPI3CON2bits.SPIROVEN = 0;	// Receive overflow does not generate error events
	SPI3CON2bits.SPITUREN = 0;	// Transmit Underrun Does Not Generates Error Events
	SPI3CON2bits.IGNROV = 0;	// A ROV is a critical error which stop SPI operation
	SPI3CON2bits.IGNTUR	= 0;	// A TUR is a critical error which stop SPI operation
	SPI3CON2bits.AUDEN = 0;		// Audio protocol disabled
	SPI3CON2bits.AUDMONO = 0;	// Audio data is stereo
	SPI3CON2bits.AUDMOD = 0b00;	// Audio protocol mode is I2S mode

	SPI3STATbits.SPIROV = 0;	// cleared Receive Overflow Flag bit. 
	SPI3CONbits.ON = 1;	// SPI Peripheral is enabled

***/
	SPI3CON = 0;
	SPI3BRG = 1;			// SCK3 = 80MHz/2/(SPI3BRG+1) = 20MHz
	SPI3CONSET = 0x8060;	// ON=1, CKP=1, MSTEN=1
	SPI3STATCLR = 0x40;		// clear SPIROV

#else
	#error "Undefined processer"
#endif

    // Init UART
    UART2Init();

	// LCDC initialize
	lcdc_initialize();

	// USB Host driver
	USBHostInit(0);


} // InitializeSystem


/*------------------------------------------------------------------------------
	convert yuyv to rgb565
------------------------------------------------------------------------------*/

void	yuv2rgb_mono(BYTE *dst, BYTE *src, WORD length)
{

	BYTE y;
	WORD n;

	// convert from YUYV to RGB565
	for(n = 0; n < length; n+=4, src+=4){

		y = *src;
		*dst++ = ((y << 3) & 0b11100000) | ((y >> 3) & 0b00011111);	// low byte
		*dst++ = ((y & 0b11111000) | ((y >> 5) & 0b00000111)); // high byte

		y = *(src+2);
		*dst++ = ((y << 3) & 0b11100000) | ((y >> 3) & 0b00011111);	// low byte
		*dst++ = ((y & 0b11111000) | ((y >> 5) & 0b00000111)); // high byte

	}
}


void	yuv2rgb_color(BYTE *dst, BYTE *src, WORD length)
{

	int y, u, v, i, j, k, l;
	int	r, g, b;
	WORD n;


	// convert from YUYV to RGB565
	for(n = 0; n < length; n+=4, src+=4){

		u = (int)(*(src+1)) - 128;
		v = (int)(*(src+3)) - 128;

		// YUV to RGB24
#if defined(__PIC32MX__)
		//  for 32 bits MICOM
		i = 359 * v;
		j = 88 * u;
		k = 183 * v;
		l = 454 * u;

		y = (int)(*src) << 8;
		r = (y + i) >> 8;
		g = (y - j - k) >> 8;
		b = (y + l) >> 8;

#else	// PIC24F 16 bits
		// YUV to RGB24
		//  for 16 bits MICOM
		i = 22 * v;
		j = 6 * u;
		k = 11 * v;
		l = 28 * u;

		y = (int)(*src) << 4;

		//r = (y + (22 * v)) >> 4;
        //g = (y - (6 * u) - (11 * v)) >> 4;
        //b = (y + (28 * u)) >> 4;

		r = (y + i) >> 4;
		g = (y - j - k) >> 4;
		b = (y + l) >> 4;
#endif

		if(r < 0){
			r = 0;
		}else if(r > 255){
			r = 255;
		}

		if(g < 0){
			g = 0;
		}else if(g > 255){
			g = 255;
		}

		if(b < 0){
			b = 0;
		}else if(b > 255){
			b = 255;
		}

		// RGB24 to RGB565
		*dst++ = ((g << 3) & 0b11100000) | ((b >> 3) & 0b00011111);	// low byte
		*dst++ = ((r & 0b11111000) | ((g >> 5) & 0b00000111)); // high byte

		// YUV to RGB24
#if defined(__PIC32MX__)
		y = (int)(*(src+2)) << 8;
		r = (y + i) >> 8;
		g = (y - j - k) >> 8;
		b = (y + l) >> 8;
#else
		y = (int)(*(src+2)) << 4;
		r = (y + i) >> 4;
		g = (y - j - k) >> 4;
		b = (y + l) >> 4;
#endif 

		if(r < 0){
			r = 0;
		}else if(r > 255){
			r = 255;
		}

		if(g < 0){
			g = 0;
		}else if(g > 255){
			g = 255;
		}

		if(b < 0){
			b = 0;
		}else if(b > 255){
			b = 255;
		}

		// RGB24 to RGB565
		*dst++ = ((g << 3) & 0b11100000) | ((b >> 3) & 0b00011111);	// low byte
		*dst++ = ((r & 0b11111000) | ((g >> 5) & 0b00000111)); // high byte

	}
}


/*------------------------------------------------------------------------------
	getstr
------------------------------------------------------------------------------*/

void	getstr(char *buf)
{
	int	c;

	while(camera_state == CAMERA_ATTACHED){
		USBHostTasks();
		if(inkey()){
			c = getch();
			if(c >= ' ' && c <= 0x7e){
				*buf++ = c;
				putch(c);       // echo back
			}else if(c == '\r'){
				putstr("\r\n");	// echo back
				break;
			}
		}
	}
	*buf = 0;
}


/*------------------------------------------------------------------------------
	putdata 
------------------------------------------------------------------------------*/

void	putdata(BYTE *data, BYTE len, WORD type)
{
	BYTE	n;
	int		i;
	char	pbuf[10];

	if(data == NULL || len == 0 || type == 0) return;

	n = 0;
	while(type){
		switch(type & 0xf){ // check low half byte
		case BOOLEAN:
			putstr(data[n]? "1 ": "0 ");
			n++;
			break;

		case bBITMAP:
			for(i = 7; i >= 0; i--){
				if(data[n] & (1<<i)) putch('1');
				else putch('0');
			}
			n++;
			break;

		case wBITMAP:
			for(i = 15; i >= 0; i--){
				if(data[n] & (1<<i)) putch('1');
				else putch('0');
			}
			n += 2;
			break;

		case bSIGNED:
			sprintf(pbuf, "%d ", (int)data[n]); putstr(pbuf);
			n++;
			break;

		case wSIGNED:
			sprintf(pbuf, "%d ", ((int)(data[n+1]) << 8) + data[n]); putstr(pbuf);
			n += 2;
			break;

		case dwSIGNED:
			sprintf(pbuf, "%ld",  
				((long)(data[n+3]) << 24) + ((long)(data[n+2]) << 16)+ ((long)(data[n+1]) << 8) + data[n]
				); putstr(pbuf);
			n += 4;
			break;

		case bNUMBER:
			sprintf(pbuf, "%u ", (WORD)data[n]); putstr(pbuf);
			n++;
			break;

		case wNUMBER:
			sprintf(pbuf, "%u ", ((WORD)(data[n+1]) << 8) + data[n]); putstr(pbuf);
			n += 2;
			break;

		case dwNUMBER:
			sprintf(pbuf, "%lu",  
				((DWORD)(data[n+3]) << 24) + ((DWORD)(data[n+2]) << 16)+ ((DWORD)(data[n+1]) << 8) + data[n]
				); putstr(pbuf);
			n += 4;
			break;

		default:
			break;
		}
		type >>= 4;	// left shift half byte
	}

}


/*-------------------------------------------------------------------------------
	BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
-------------------------------------------------------------------------------*/

BOOL USB_ApplicationEventHandler ( BYTE address, USB_EVENT event, void *data, DWORD size )
{

    // Handle specific events.
    switch (event){
    case EVENT_UVC_ATTACH:
		camera_state = CAMERA_ATTACHED;
        putstr("Camera attached.\r\n");
		break;

    case EVENT_UVC_DETACH:
    	camera_state = CAMERA_DETACHED;
        putstr( "Camera detached.\r\n" );
		break;

    default:
       break;
   }
   return TRUE;

} // USB_ApplicationEventHandler


/**** end of main.c ************************************************************/
